from pyspark.sql import SparkSession, DataFrame, Window
from pyspark.sql.functions import row_number, col, date_format, to_timestamp, udf, when

from common.constants import SRC_PATH_USER, USER_DIM, CALENDAR_DIM, TRACK_DIM
from common.functions import valid_timestamp


class StreamTransformFunctions:
    def __init__(self, spark, stream):
        self.stream: DataFrame = stream
        self.spark: SparkSession = spark

    def compute_stream_duration(self):
        self.stream = self.stream.withColumn("action_number",
                                             row_number().over(Window
                                                               .partitionBy("email", "track", "artist", "album")
                                                               .orderBy("datetime")))

        self.stream = self.stream.alias('play') \
            .join(self.stream.alias('pause'),
                  [col('play.track') == col('pause.track'), col('play.artist') == col('pause.artist'),
                   col('play.album') == col('pause.album'), col('play.email') == col('pause.email')],
                  'left') \
            .where((col('play.action') == 'play') & (col('pause.action') == 'pause') &
                   (col('pause.action_number') == col('play.action_number') + 1)) \
            .select(date_format(col('play.datetime'), 'yyyy-MM-dd').alias('date'),
                    'play.email', 'play.track', 'play.artist', 'play.album',
                    to_timestamp(col('play.datetime'), 'yyyy-MM-dd HH:mm:ss').alias('play_time'),
                    to_timestamp(col('pause.datetime'), 'yyyy-MM-dd HH:mm:ss').alias('pause_time')) \
            .withColumn("stream_duration",
                        (col("pause_time").cast("long") - col("play_time").cast("long"))) \
            .drop('pause_time', 'play_time')

    def validate(self):
        valid_timestamp_udf = udf(lambda x: valid_timestamp(x))
        self.stream = self.stream.withColumn('datetime', valid_timestamp_udf(col('datetime').cast('string')))

    def eliminate_nulls(self):
        self.stream = self.stream.where((self.stream['datetime'] != 'unknown') &
                                        (self.stream['email'].isNotNull()) &
                                        (self.stream['track'].isNotNull()) &
                                        (self.stream['artist'].isNotNull()) &
                                        (self.stream['album'].isNotNull()))

    def add_premium_dim(self):
        user_src = self.spark.read.load(SRC_PATH_USER, format='csv', sep=",", header="true", inferSchema="true")

        self.stream = self.stream \
            .join(user_src,
                  self.stream['email'] == user_src['email'],
                  'left') \
            .select(self.stream['*'], user_src['premium']) \
            .withColumn('premium',
                        when(col('premium').cast('string') == 'true', 'premium')
                        .when(col('premium').cast('string') == 'false', 'basic')
                        .otherwise('unknown'))

    def assign_sk(self):
        calendar_dim = self.spark.table(CALENDAR_DIM)
        self.stream = self.stream.join(calendar_dim,
                                       self.stream['date'] == calendar_dim['date'],
                                       'left'
                                       ).select(calendar_dim['date_id'], self.stream['*']
                                                ).drop('date')
        self.stream = self.stream.fillna({'date_id': -1})

        user_dim = self.spark.table(USER_DIM)
        self.stream = self.stream.join(user_dim,
                                       self.stream['email'] == user_dim['email'],
                                       'left'
                                       ).select(user_dim['user_id'], self.stream['*']
                                                ).drop('email')
        self.stream = self.stream.fillna({'user_id': -1})

        track_dim = self.spark.table(TRACK_DIM)
        self.stream = self.stream.join(track_dim,
                                       [self.stream['track'] == track_dim['title'],
                                        self.stream['artist'] == track_dim['artist'],
                                        self.stream['album'] == track_dim['album']],
                                       'left'
                                       ).select(track_dim['track_id'], self.stream['*']
                                                ).drop('track', 'artist', 'album')
        self.stream = self.stream.fillna({'track_id': -1})
        return self.stream
