from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import col, udf

from common.constants import TRACK_DIM
from common.functions import create_sk, valid_date


class TrackTransformFunctions:
    def __init__(self, spark, track, album):
        self.track: DataFrame = track
        self.album: DataFrame = album
        self.spark: SparkSession = spark

    def get_null_row(self):
        columns = ['track_id', 'title', 'artist', 'album', 'genre', 'duration', 'release_date', 'publisher']
        vals = [(-1,) + (('unknown',) * 7)]
        return self.spark.createDataFrame(vals, columns)

    def eliminate_nulls(self):
        self.track = self.track.where((self.track['title'].isNotNull()) &
                                      (self.track['artist'].isNotNull()) &
                                      (self.track['album'].isNotNull()))
        self.track = self.track.fillna('unknown')

    def validate(self):
        valid_date_udf = udf(lambda x: valid_date(x))
        self.track = self.track.withColumn('release_date', valid_date_udf(col('release_date')))

    def join_albums(self):
        self.track = self.track.join(self.album,
                                     [self.track['artist'] == self.album['artist'],
                                      self.track['album'] == self.album['title']],
                                     'left').select(self.track['title'],
                                                    self.track['artist'],
                                                    self.track['album'],
                                                    self.track['genre'],
                                                    self.track['duration_ms'].cast('string').alias('duration'),
                                                    self.album['release_date'].cast('string'),
                                                    self.album['publisher'])

    def prep_init_form(self):
        self.track = create_sk(self.spark, self.track, 1000001, 'track_id')

        self.track = self.get_null_row().union(self.track)

        return self.track

    def prep_scd_form(self):
        track_dim = self.spark.table(TRACK_DIM)

        diff_table = self.track \
            .join(track_dim,
                  [self.track['title'] == track_dim['title'],
                   self.track['album'] == track_dim['album'],
                   self.track['artist'] == track_dim['artist']],
                  'left') \
            .where((track_dim['track_id'].isNull()) |
                   (self.track['genre'] != track_dim['genre']) |
                   (self.track['duration'] != track_dim['duration']) |
                   (self.track['release_date'] != track_dim['release_date']) |
                   (self.track['publisher'] != track_dim['publisher']))

        diff_table = diff_table.select(track_dim['track_id'],
                                       self.track['*'])

        # NEW ROWS
        new_rows = diff_table.where(diff_table['track_id'].isNull()).drop('track_id')
        id_offset = track_dim.agg({"track_id": "max"}).collect()[0][0]
        new_rows = create_sk(self.spark, new_rows, id_offset + 1, "track_id")

        # SCD
        changed_rows = diff_table.where(diff_table['track_id'].isNotNull())

        self.track = track_dim \
            .join(changed_rows, track_dim['track_id'] == changed_rows['track_id'], 'leftanti') \
            .union(changed_rows).union(new_rows)

        return self.track
