from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import when, col, lit, udf

from common.constants import EXT_PATH_GEO_DATA, USER_DIM
from common.functions import create_sk, valid_date


class UserTransformFunctions:
    def __init__(self, spark, user):
        self.user: DataFrame = user
        self.spark: SparkSession = spark

    def get_null_row(self):
        columns = ['user_id', 'email', 'firstname', 'gender', 'birthday', 'country', 'region', 'is_current']
        vals = [(-1,) + (('unknown',) * 6) + (True,)]
        return self.spark.createDataFrame(vals, columns)

    def eliminate_nulls(self):
        self.user = self.user.where(self.user['email'].isNotNull()).fillna('unknown')
        self.user = self.user.where(self.user['region'].isNotNull()).fillna('unknown')

    def validate(self):
        valid_date_udf = udf(lambda x: valid_date(x))
        self.user = self.user.withColumn('birthday', valid_date_udf(col('birthday')))

    def translate_gender(self):
        self.user = self.user.withColumn('gender',
                                         when(col('gender') == 1, 'male')
                                         .when(col('gender') == 2, 'female')
                                         .otherwise('unknown'))

    def assign_region(self):
        geo_data = self.spark.read.load(EXT_PATH_GEO_DATA, format='csv', sep=",", header="true", inferSchema="true")

        self.user = self.user.join(geo_data,
                                   [self.user['country'] == geo_data['country_name']],
                                   'left').select(self.user['email'],
                                                  self.user['name'].alias('firstname'),
                                                  self.user['gender'],
                                                  self.user['birthday'].cast('string'),
                                                  self.user['country'],
                                                  geo_data['continent'].alias('region'))

    def prep_init_form(self):
        self.user = create_sk(self.spark, self.user, 1000001, 'user_id')

        self.user = self.user.withColumn('is_current', lit(True))

        self.user = self.get_null_row().union(self.user)

        return self.user

    def prep_scd_form(self):
        user_dim = self.spark.table(USER_DIM)

        diff_table = self.user \
            .join(user_dim,
                  self.user['email'] == user_dim['email'],
                  'left') \
            .where((user_dim['user_id'].isNull()) |
                   (self.user['gender'] != user_dim['gender']) |
                   (self.user['birthday'] != user_dim['birthday']) |
                   ((user_dim['is_current']) & (self.user['country'] != user_dim['country']))) \
            .select(user_dim['user_id'],
                    self.user['*'],
                    user_dim['country'].alias('country_'),
                    user_dim['region'].alias('region_'),
                    user_dim['is_current'])

        # NEW ROWS
        new_rows = diff_table.where(diff_table['user_id'].isNull()) \
            .select(diff_table['email'],
                    diff_table['firstname'],
                    diff_table['gender'],
                    diff_table['birthday'],
                    diff_table['country'],
                    diff_table['region']) \
            .withColumn('is_current', lit(True))

        # generate IDs
        id_offset = user_dim.agg({"user_id": "max"}).collect()[0][0]
        new_rows = create_sk(self.spark, new_rows, id_offset + 1, "user_id")
        id_offset += new_rows.count()

        # add new rows
        user_dim = user_dim.union(new_rows)

        # SCD
        changed_rows = diff_table.where(diff_table['user_id'].isNotNull())

        new_versions = changed_rows \
            .where((changed_rows['country'] != changed_rows['country_']) &
                   (changed_rows['is_current'])) \
            .drop('country_', 'region_', 'user_id') \
            .withColumn('is_current', lit(True))
        new_versions = create_sk(self.spark, new_versions, id_offset + 1, 'user_id')

        changed_rows = changed_rows \
            .withColumn('is_current',
                        when(changed_rows['country'] == changed_rows['country_'],
                             changed_rows['is_current']).otherwise(lit(False))) \
            .drop('country', 'region') \
            .withColumnRenamed('country_', 'country') \
            .withColumnRenamed('region_', 'region')

        changed_rows = changed_rows.union(new_versions)

        self.user = user_dim \
            .join(changed_rows, user_dim['user_id'] == changed_rows['user_id'], 'leftanti') \
            .union(changed_rows)

        return self.user
